package tools.descartes.dql.connector.kieker.filter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.InputPort;
import kieker.analysis.plugin.annotation.OutputPort;
import kieker.analysis.plugin.annotation.Plugin;
import kieker.common.configuration.Configuration;
import kieker.tools.traceAnalysis.filter.AbstractMessageTraceProcessingFilter;
import kieker.tools.traceAnalysis.systemModel.AbstractMessage;
import kieker.tools.traceAnalysis.systemModel.MessageTrace;
import kieker.tools.traceAnalysis.systemModel.Operation;
import tools.descartes.dql.connector.kieker.structures.Session;

import org.apache.log4j.Logger;

@Plugin(name = "SessionFilter", description = "...", outputPorts = {
		@OutputPort(name = SessionFilter.OUTPUT_PORT_NAME_WORKLOAD, description = "Outputs workload description", eventTypes = {
				HashMap.class }), })
public class SessionFilter extends AbstractMessageTraceProcessingFilter {

	public static final String INPUT_PORT_NAME_MESSAGE_TRACE = "messageTrace";
	public static final String OUTPUT_PORT_NAME_WORKLOAD = "workload";

	private static final Logger log = Logger.getLogger(SessionFilter.class);
	private static HashMap<String, List<Long>> timeSeriesMap;
	private static HashMap<String, Long> sumMap;
	public static ArrayList<Session> sessionTimes;

	public SessionFilter(Configuration configuration, IProjectContext projectContext) {
		super(configuration, projectContext);
		timeSeriesMap = new HashMap<String, List<Long>>();
		sumMap = new HashMap<String, Long>();
		sessionTimes = new ArrayList<Session>();
	}

	@Override
	@InputPort(name = AbstractMessageTraceProcessingFilter.INPUT_PORT_NAME_MESSAGE_TRACES, description = "Receives the message traces to be processed", eventTypes = {
			MessageTrace.class })
	public void inputMessageTraces(final MessageTrace mt) {
		long start = mt.getStartTimestamp();
		long time = (long) ((mt.getEndTimestamp() - mt.getStartTimestamp()) * 0.000001);
		// mt.getTraceInformation().getSessionId();
		AbstractMessage message = getStartMessage(mt);
		long time2 = message.getReceivingExecution().getTin();
		// double time3 = message.getReceivingExecution().getTout();
		long time3 = message.getReceivingExecution().getTout();

		Operation operation = message.getReceivingExecution().getOperation();
		String id = mt.getSessionId();
		System.out.println(id + ":" + time + "\txxx" + (time3 - time2) * 0.000001 + "\t\t "
				+ operation.getSignature().getName() + " \t\t||| " + start);

		// addTimeStamp(id, mt.getStartTimestamp());
		// addTimeStamp(id, mt.getEndTimestamp());
		// addTimeStamp(id, message.getTimestamp());
		addTimeStamp(id, message.getReceivingExecution().getTin());
		addTimeStamp(id, message.getReceivingExecution().getTout());
		long value = (time3 - time2); //* 0.000001;
		System.out.println("\t\t\t"+value);
		addToSumMap(id, value);
		// addTimeStamp(id, message.getReceivingExecution().getTout());
		sortMessageTrace(mt);
	}
	Comparator<AbstractMessage> timeComparator;
	
	private void sortMessageTrace(final MessageTrace mt) {
		List<AbstractMessage> messages = mt.getSequenceAsVector();
		messages = new ArrayList<AbstractMessage>(messages);
		timeComparator = new Comparator<AbstractMessage>() {
			@Override
			public int compare(AbstractMessage arg0, AbstractMessage arg1) {
				return arg0.getReceivingExecution().getEoi() - arg1.getReceivingExecution().getEoi();
//				if (arg0.getTimestamp() > arg1.getTimestamp()) {
//					return 1;
//				} else if (arg0.getTimestamp() < arg1.getTimestamp()) {
//					return -1;
//				} else {
//					return 0;
//				}
			}
		};
		Collections.sort(messages, timeComparator);
		for (AbstractMessage message : messages) {
			if(message.getReceivingExecution().getEoi() == 0){
				System.out.println(
					"==>" + message.getReceivingExecution().getOperation().getSignature().getName() + " " + message.getReceivingExecution().getEoi());
			}else{
				System.out.println(
						"" + message.getReceivingExecution().getOperation().getSignature().getName() + " " + message.getReceivingExecution().getEoi());
			}
		}
	}

	AbstractMessage getStartMessage(final MessageTrace mt) {
		for (AbstractMessage message : mt.getSequenceAsVector()) {
			System.out.println("\t" + message.getReceivingExecution().getOperation().getSignature().getName() + " "
					+ message.getTimestamp() + " " + mt.getStartTimestamp());
			// System.out.println("\t"+message.getReceivingExecution().getTin()
			// + " "+ message.getReceivingExecution().getTout() + " " +
			// message.getSendingExecution().getTout() );
			if (message.getTimestamp() == mt.getStartTimestamp()) {
				return message;
			}
		}
		return mt.getSequenceAsVector().get(0);
	}

	private synchronized void addToSumMap(String key, long time){
		if (!sumMap.containsKey(key)) {
			sumMap.put(key, time);
		} else {
			sumMap.put(key, sumMap.get(key)+ time);
		}
	}

	private synchronized void addTimeStamp(String key, long timestamp) {
		List<Long> timeSeries;
		if (!timeSeriesMap.containsKey(key)) {
			timeSeries = new ArrayList<>();
			timeSeries.add(timestamp);
			timeSeriesMap.put(key, timeSeries);
		} else {
			timeSeries = timeSeriesMap.get(key);
			timeSeries.add(timestamp);
		}
	}

	@Override
	public void terminate(boolean errorBeforeTermination) {
		// log.info("workloadddd "+timeSeriesMap);
		if (timeSeriesMap.isEmpty()) {
			log.error("Could not extract incoming requests to create workload model");
		}

		for (String key : timeSeriesMap.keySet()) {
			List<Long> list = timeSeriesMap.get(key);
			Collections.sort(timeSeriesMap.get(key));
			for (long t : list) {
				//System.out.println("\t\t" + (t * 0.000001));
			}
			;
			long start = Collections.min(list);//list.get(0);
			long duration = (long) (Collections.max(list) - start);
//			System.out.println(key + "\t\t" + time + "\t\t\tnumElements:"+list.size()+"\tfirst:" + list.get(0) + "\tlast" + list.get(list.size() - 1));
//			System.out.println(key + "\t\t" +sumMap.get(key)+ " << sum map");
			//System.out.println(key + "\t\t" + time + "<< total \t\t" + sumMap.get(key) * 0.000001 + " << processing time");
			addNewSession(key, start, duration);
		}

		super.deliver(SessionFilter.OUTPUT_PORT_NAME_WORKLOAD, timeSeriesMap);
		super.terminate(errorBeforeTermination);
	}

	private void addNewSession(String key, long start, long duration) {
		sessionTimes.add(new Session(key, start, duration));
	}

	@Override
	public Configuration getCurrentConfiguration() {
		return this.configuration;
	}

}
